#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""blamlib.net is a nicer design then the orignal blamlib.network
since it do not require async sockets.

It is split into two parts (or three but that dependse on how you count).

Workers - "low level" network communications, handles tcp/upd/whatever
and sendss events to the protocols. These run in a thread.

Clients - small classes that should be used for interating with the
current worker, these are automatically created.

Protocols - well defines what to do with the traffic.

A simple application could look like this:

net = UDP_Worker( EchoProtocol, 'localhost', 5600, dict() )
net.start()

while net.work:
   time.sleep( 1 )"""

import socket, threading, errno


#--- UDP ---
class UDP_Worker( threading.Thread ):
    def __init__( self, protocol_class, host, port, protocol_args ):
        threading.Thread.__init__( self )
        self.send_lock = threading.Lock()
        self.protocol = protocol_class( self, protocol_args )
        self.work = True
        self.sock_readsize = 1024
        self.host = host
        self.port = port
        self.clients = list()
        
    def run( self ):
        # setup network
        self.sock = socket.socket( socket.AF_INET, socket.SOCK_DGRAM )
        self.sock.setsockopt( socket.SOL_SOCKET, socket.SO_REUSEADDR, True )
        self.sock.bind( ( self.host, self.port ) )
        
        # do a funny loop and try to read data
        while self.work:          
            # read new data
            data, from_addr = self.sock.recvfrom( self.sock_readsize )
            cl_hash = hash( from_addr )
            
            # hack (must be a better way for udp..
            if cl_hash not in map( lambda client: client.hash, self.clients ):
                self.clients.append(
                    UDP_Client( self, from_addr[0], from_addr[1] ) )
            
            # handle new data
            client = filter( lambda x: x.hash == cl_hash, self.clients )[0]
            self.protocol.handle( client, data )

        # disconnect network

    def sendtoall( self, data ):
        for client in self.clients:
            client.send( data )

    def stop( self ):
        self.sock.shutdown( socket.SHUT_RDWR )
        self.sock.close()
        self.work = False

class UDP_Client:
    def __init__( self, worker, host, port ):
        self.worker = worker
        self.host = host
        self.port = port
        self.hash = hash( ( host, port ) )

    def send( self, data ):
        self.worker.send_lock.acquire() # I think this is a great idea.
        self.worker.sock.sendto( data, ( self.host, self.port ) )
        self.worker.send_lock.release()

    def close( self ):
        # More logical for TCP...
        pass


# --- TCP ---
class TCP_Worker( UDP_Worker ):
    def __init__( self, protocol_class, host, port, protocol_args ):
        threading.Thread.__init__( self )
        self.send_lock = threading.Lock()
        self.protocol = protocol_class( self, protocol_args )
        self.work = True
        self.sock_readsize = 1024
        self.host = host
        self.port = port
        self.clients = list()
        
    def run( self ):
        # setup network
        self.sock = socket.socket( socket.AF_INET, socket.SOCK_STREAM )
        self.sock.setsockopt( socket.SOL_SOCKET, socket.SO_REUSEADDR, True )
        self.sock.bind( ( self.host, self.port ) )
        self.sock.listen( True )
        
        # do a funny loop and try to read data
        while self.work:
            # listen for new connections...
            try:
                sock_stuff = self.sock.accept()
            except socket.error, why:
                if why[0] in ( errno.EINVAL, ):
                    self.work = False # Shutdown
                    break
                else:
                    raise socket.error, why
            
            client = TCP_Client( self, sock_stuff )
            client.start()
            self.clients.append( client )

            # Filter out all sockets that are no longer active.
            self.clients == filter( lambda x: x.online == True, self.clients )

        # disconnect network
        self.stop() # can't be that bad...
        
    def stop( self ):
        for client in self.clients:
            client.close()

        try:
            self.sock.shutdown( socket.SHUT_RDWR )
            self.sock.close()
        except socket.error, why:
            if why[0] in ( errno.EBADF, ):
                pass # no real problem... wonder why it happens really..
            else:
                raise socket.error, why

        self.work = False

    def close_client( self, client ):
        possible = filter( lambda c: c.hash == client.hash, self.clients )

        if len( possible ) != 1:
            raise ValueError, "Tried to close a non-existent client"

        possible[0].close()
        
class TCP_Client( threading.Thread ):
    def __init__( self, worker, socket ):
        threading.Thread.__init__( self )
        self.sock = socket[0]
        self.host = socket[1][0]
        self.port = socket[1][1]
        self.hash = hash( socket[1] )
        self.worker = worker
        self.online = True
        self.worker.protocol.on_connect( self )

    def run( self ):
        while self.online:
            try:
                buff = self.sock.recv( self.worker.sock_readsize )
                self.worker.protocol.handle( self, buff )

            except socket.error, why:
                if why[0] in ( errno.ECONNRESET, errno.ENOTCONN,
                               errno.ETIMEDOUT, errno.EPIPE ):
                    self.close()
                    return # that is kill the thread :)
                else:
                    raise socket.error, why

    def send( self, data ):
        #self.worker.send_lock.acquire() # maybe not a necessary?
        self.sock.sendall( data )
        #self.worker.send_lock.release()

    def close( self ):
        # magic here
        self.worker.protocol.on_disconnect( self )
        try:
            self.sock.shutdown( socket.SHUT_RDWR )
            self.sock.close()
        except socket.error, why:
            if why[0] in ( errno.EBADF, errno.ENOTCONN ):
                pass # closing a dead client :)
            else:
                raise socket.error, why
        
        self.online = False



# -- FIFO
class Fifo_Worker( UDP_Worker ):
    """Yea use a FIFO as a network transport"""

    def __init__( self, protocol_class, host, port, protocol_args ):
        """Okey now Fifo_Worker is a bit strange, so just specify the
        FIFO filename as host, and set port to something."""
        threading.Thread.__init__( self )
        self.protocol = protocol_class( self, protocol_args )
        self.work = True
        self.sock_readsize = "line" # 1024
        self.host = host
        self.port = port
        self.client = self.Fifo_Client( self )
        self.in_fifo = open( self.host, 'r' )
        self.out_fifo = open( self.host, 'w' )
        
    def run( self ):
        
        while self.work:
            if type( self.sock_readsize ) == int:
                data = self.in_fifo.read( self.sock_readsize )
            else:
                data = self.in_fifo.readline()
            self.protocol.handle( self.client, data )

        self.in_fifo.close()
        self.out_fifo.close()

    def sendtoall( self, data ):
        # ehm.. yea...
        self.out_fifo.write( data )
        self.out_fifo.flush()
        
    def stop( self ):
        self.work = False

    class Fifo_Client:
        def __init__( self, worker ):
            self.worker = worker
            self.host = worker.host
            self.port = worker.port

        def send( self, data ):
            self.worker.sendtoall( data )

        def stop( self ):
            pass


# --- Sample protocol ---
class Protocol:
    def __init__( self, worker, args ):
        self.net = worker

    def on_connect( self, client ):
        """On connect (new client)"""
        print "%s:d Connected" %( client.host, client.port )

    def on_disconnect( self, client ):
        """Called when a client disconnected"""
        print "%s:d Disconnected" %( client.host, client.port )

    def handle( self, client, data ):
        """Called when we got data from a client"""
        print "<%s:%d> '%s'" %( client.host, client.port , data.strip() )
        client.send( data )
